# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qwak/feature_store/v1/internal/data_source/data_source_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nDqwak/feature_store/v1/internal/data_source/data_source_service.proto\x12\x31qwak.feature.store.management.v1.internal.sources\"E\n)GetDataSourceSourceCodeDownloadURLRequest\x12\x18\n\x10\x64\x61ta_source_name\x18\x01 \x01(\t\"B\n*GetDataSourceSourceCodeDownloadURLResponse\x12\x14\n\x0c\x64ownload_url\x18\x01 \x01(\t\"@\n$GetDataSourceSourceCodeUploadRequest\x12\x18\n\x10\x64\x61ta_source_name\x18\x01 \x01(\t\";\n%GetDataSourceSourceCodeUploadResponse\x12\x12\n\nupload_url\x18\x01 \x01(\t2\xcf\x03\n\x11\x44\x61taSourceService\x12\xd5\x01\n GetDataSourceSourceCodeUploadURL\x12W.qwak.feature.store.management.v1.internal.sources.GetDataSourceSourceCodeUploadRequest\x1aX.qwak.feature.store.management.v1.internal.sources.GetDataSourceSourceCodeUploadResponse\x12\xe1\x01\n\"GetDataSourceSourceCodeDownloadURL\x12\\.qwak.feature.store.management.v1.internal.sources.GetDataSourceSourceCodeDownloadURLRequest\x1a].qwak.feature.store.management.v1.internal.sources.GetDataSourceSourceCodeDownloadURLResponseBw\n<com.qwak.ai.feature.store.management.v1.api.internal.sourcesP\x01Z5qwak/featurestore/sources;featurestoreinternalsourcesb\x06proto3')



_GETDATASOURCESOURCECODEDOWNLOADURLREQUEST = DESCRIPTOR.message_types_by_name['GetDataSourceSourceCodeDownloadURLRequest']
_GETDATASOURCESOURCECODEDOWNLOADURLRESPONSE = DESCRIPTOR.message_types_by_name['GetDataSourceSourceCodeDownloadURLResponse']
_GETDATASOURCESOURCECODEUPLOADREQUEST = DESCRIPTOR.message_types_by_name['GetDataSourceSourceCodeUploadRequest']
_GETDATASOURCESOURCECODEUPLOADRESPONSE = DESCRIPTOR.message_types_by_name['GetDataSourceSourceCodeUploadResponse']
GetDataSourceSourceCodeDownloadURLRequest = _reflection.GeneratedProtocolMessageType('GetDataSourceSourceCodeDownloadURLRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASOURCESOURCECODEDOWNLOADURLREQUEST,
  '__module__' : 'qwak.feature_store.v1.internal.data_source.data_source_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.management.v1.internal.sources.GetDataSourceSourceCodeDownloadURLRequest)
  })
_sym_db.RegisterMessage(GetDataSourceSourceCodeDownloadURLRequest)

GetDataSourceSourceCodeDownloadURLResponse = _reflection.GeneratedProtocolMessageType('GetDataSourceSourceCodeDownloadURLResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASOURCESOURCECODEDOWNLOADURLRESPONSE,
  '__module__' : 'qwak.feature_store.v1.internal.data_source.data_source_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.management.v1.internal.sources.GetDataSourceSourceCodeDownloadURLResponse)
  })
_sym_db.RegisterMessage(GetDataSourceSourceCodeDownloadURLResponse)

GetDataSourceSourceCodeUploadRequest = _reflection.GeneratedProtocolMessageType('GetDataSourceSourceCodeUploadRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASOURCESOURCECODEUPLOADREQUEST,
  '__module__' : 'qwak.feature_store.v1.internal.data_source.data_source_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.management.v1.internal.sources.GetDataSourceSourceCodeUploadRequest)
  })
_sym_db.RegisterMessage(GetDataSourceSourceCodeUploadRequest)

GetDataSourceSourceCodeUploadResponse = _reflection.GeneratedProtocolMessageType('GetDataSourceSourceCodeUploadResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASOURCESOURCECODEUPLOADRESPONSE,
  '__module__' : 'qwak.feature_store.v1.internal.data_source.data_source_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.management.v1.internal.sources.GetDataSourceSourceCodeUploadResponse)
  })
_sym_db.RegisterMessage(GetDataSourceSourceCodeUploadResponse)

_DATASOURCESERVICE = DESCRIPTOR.services_by_name['DataSourceService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n<com.qwak.ai.feature.store.management.v1.api.internal.sourcesP\001Z5qwak/featurestore/sources;featurestoreinternalsources'
  _GETDATASOURCESOURCECODEDOWNLOADURLREQUEST._serialized_start=123
  _GETDATASOURCESOURCECODEDOWNLOADURLREQUEST._serialized_end=192
  _GETDATASOURCESOURCECODEDOWNLOADURLRESPONSE._serialized_start=194
  _GETDATASOURCESOURCECODEDOWNLOADURLRESPONSE._serialized_end=260
  _GETDATASOURCESOURCECODEUPLOADREQUEST._serialized_start=262
  _GETDATASOURCESOURCECODEUPLOADREQUEST._serialized_end=326
  _GETDATASOURCESOURCECODEUPLOADRESPONSE._serialized_start=328
  _GETDATASOURCESOURCECODEUPLOADRESPONSE._serialized_end=387
  _DATASOURCESERVICE._serialized_start=390
  _DATASOURCESERVICE._serialized_end=853
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qwak/admiral/secret/v0/system_secret_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from _qwak_proto.qwak.admiral.secret.v0 import secret_pb2 as qwak_dot_admiral_dot_secret_dot_v0_dot_secret__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2qwak/admiral/secret/v0/system_secret_service.proto\x12\x16qwak.admiral.secret.v0\x1a#qwak/admiral/secret/v0/secret.proto\"d\n\x16SetSystemSecretRequest\x12J\n\x07options\x18\x01 \x01(\x0b\x32\x39.qwak.admiral.secret.v0.SetSystemEnvironmentSecretOptions\"\x19\n\x17SetSystemSecretResponse\"d\n\x19\x44\x65leteSystemSecretRequest\x12G\n\nidentifier\x18\x01 \x01(\x0b\x32\x33.qwak.admiral.secret.v0.EnvironmentSecretIdentifier\"\x1c\n\x1a\x44\x65leteSystemSecretResponse\"a\n\x16GetSystemSecretRequest\x12G\n\nidentifier\x18\x01 \x01(\x0b\x32\x33.qwak.admiral.secret.v0.EnvironmentSecretIdentifier\"d\n\x17GetSystemSecretResponse\x12I\n\x11secret_definition\x18\x01 \x01(\x0b\x32..qwak.admiral.secret.v0.SystemSecretDefinition2\xfa\x02\n\x13SystemSecretService\x12r\n\x0fSetSystemSecret\x12..qwak.admiral.secret.v0.SetSystemSecretRequest\x1a/.qwak.admiral.secret.v0.SetSystemSecretResponse\x12{\n\x12\x44\x65leteSystemSecret\x12\x31.qwak.admiral.secret.v0.DeleteSystemSecretRequest\x1a\x32.qwak.admiral.secret.v0.DeleteSystemSecretResponse\x12r\n\x0fGetSystemSecret\x12..qwak.admiral.secret.v0.GetSystemSecretRequest\x1a/.qwak.admiral.secret.v0.GetSystemSecretResponseB\x8c\x01\n\x1d\x63om.qwak.ai.admiral.secret.v0P\x01Zigithub.com/qwak-ai/qwak-platform/services/core/go/admiral/admiral-api/pb/qwak/admiral/secret/v0;secret_v0b\x06proto3')



_SETSYSTEMSECRETREQUEST = DESCRIPTOR.message_types_by_name['SetSystemSecretRequest']
_SETSYSTEMSECRETRESPONSE = DESCRIPTOR.message_types_by_name['SetSystemSecretResponse']
_DELETESYSTEMSECRETREQUEST = DESCRIPTOR.message_types_by_name['DeleteSystemSecretRequest']
_DELETESYSTEMSECRETRESPONSE = DESCRIPTOR.message_types_by_name['DeleteSystemSecretResponse']
_GETSYSTEMSECRETREQUEST = DESCRIPTOR.message_types_by_name['GetSystemSecretRequest']
_GETSYSTEMSECRETRESPONSE = DESCRIPTOR.message_types_by_name['GetSystemSecretResponse']
SetSystemSecretRequest = _reflection.GeneratedProtocolMessageType('SetSystemSecretRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETSYSTEMSECRETREQUEST,
  '__module__' : 'qwak.admiral.secret.v0.system_secret_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.admiral.secret.v0.SetSystemSecretRequest)
  })
_sym_db.RegisterMessage(SetSystemSecretRequest)

SetSystemSecretResponse = _reflection.GeneratedProtocolMessageType('SetSystemSecretResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETSYSTEMSECRETRESPONSE,
  '__module__' : 'qwak.admiral.secret.v0.system_secret_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.admiral.secret.v0.SetSystemSecretResponse)
  })
_sym_db.RegisterMessage(SetSystemSecretResponse)

DeleteSystemSecretRequest = _reflection.GeneratedProtocolMessageType('DeleteSystemSecretRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETESYSTEMSECRETREQUEST,
  '__module__' : 'qwak.admiral.secret.v0.system_secret_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.admiral.secret.v0.DeleteSystemSecretRequest)
  })
_sym_db.RegisterMessage(DeleteSystemSecretRequest)

DeleteSystemSecretResponse = _reflection.GeneratedProtocolMessageType('DeleteSystemSecretResponse', (_message.Message,), {
  'DESCRIPTOR' : _DELETESYSTEMSECRETRESPONSE,
  '__module__' : 'qwak.admiral.secret.v0.system_secret_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.admiral.secret.v0.DeleteSystemSecretResponse)
  })
_sym_db.RegisterMessage(DeleteSystemSecretResponse)

GetSystemSecretRequest = _reflection.GeneratedProtocolMessageType('GetSystemSecretRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSYSTEMSECRETREQUEST,
  '__module__' : 'qwak.admiral.secret.v0.system_secret_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.admiral.secret.v0.GetSystemSecretRequest)
  })
_sym_db.RegisterMessage(GetSystemSecretRequest)

GetSystemSecretResponse = _reflection.GeneratedProtocolMessageType('GetSystemSecretResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSYSTEMSECRETRESPONSE,
  '__module__' : 'qwak.admiral.secret.v0.system_secret_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.admiral.secret.v0.GetSystemSecretResponse)
  })
_sym_db.RegisterMessage(GetSystemSecretResponse)

_SYSTEMSECRETSERVICE = DESCRIPTOR.services_by_name['SystemSecretService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.qwak.ai.admiral.secret.v0P\001Zigithub.com/qwak-ai/qwak-platform/services/core/go/admiral/admiral-api/pb/qwak/admiral/secret/v0;secret_v0'
  _SETSYSTEMSECRETREQUEST._serialized_start=115
  _SETSYSTEMSECRETREQUEST._serialized_end=215
  _SETSYSTEMSECRETRESPONSE._serialized_start=217
  _SETSYSTEMSECRETRESPONSE._serialized_end=242
  _DELETESYSTEMSECRETREQUEST._serialized_start=244
  _DELETESYSTEMSECRETREQUEST._serialized_end=344
  _DELETESYSTEMSECRETRESPONSE._serialized_start=346
  _DELETESYSTEMSECRETRESPONSE._serialized_end=374
  _GETSYSTEMSECRETREQUEST._serialized_start=376
  _GETSYSTEMSECRETREQUEST._serialized_end=473
  _GETSYSTEMSECRETRESPONSE._serialized_start=475
  _GETSYSTEMSECRETRESPONSE._serialized_end=575
  _SYSTEMSECRETSERVICE._serialized_start=578
  _SYSTEMSECRETSERVICE._serialized_end=956
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass, field
from datetime import datetime

from _qwak_proto.qwak.models.models_pb2 import Model as ModelProto
from google.protobuf.timestamp_pb2 import Timestamp


@dataclass
class Model:
    model_id: str = field(default=None)
    uuid: str = field(default=None)
    model_name: str = field(default=None)
    model_description: str = field(default=None)
    project_id: str = field(default=None)
    created_by: str = field(default="")
    created_at: datetime = field(default=datetime.now())
    last_modified_by: str = field(default="")
    last_modified_at: datetime = field(default=datetime.now())

    def to_proto(self):
        timestamp = Timestamp()
        return ModelProto(
            model_id=self.model_id,
            uuid=self.uuid,
            display_name=self.model_name,
            model_description=self.model_description,
            project_id=self.project_id,
            created_by=self.created_by,
            created_at=timestamp.FromDatetime(self.created_at),
            last_modified_by=self.last_modified_by,
            last_modified_at=timestamp.FromDatetime(self.last_modified_at),
        )

    @staticmethod
    def from_proto(model_proto: ModelProto):
        return Model(
            model_id=model_proto.model_id,
            uuid=model_proto.uuid,
            model_name=model_proto.display_name,
            model_description=model_proto.model_description,
            project_id=model_proto.project_id,
            created_by=model_proto.created_by,
            created_at=datetime.fromtimestamp(
                model_proto.created_at.seconds + model_proto.created_at.nanos / 1e9
            ),
            last_modified_by=model_proto.last_modified_by,
            last_modified_at=datetime.fromtimestamp(
                model_proto.last_modified_at.seconds
                + model_proto.last_modified_at.nanos / 1e9
            ),
        )



import logging
import math
from io import StringIO
from typing import TYPE_CHECKING, List, Optional, Sequence, Tuple

from _qwak_proto.qwak.ecosystem.v0.ecosystem_pb2 import AuthenticatedUnifiedUserContext
from _qwak_proto.qwak.feature_store.serving.serving_pb2 import (
    BatchFeature as ServingProtoBatchFeature,
    BatchV1Feature as ServingProtoBatchV1Feature,
    EntitiesHeader,
    EntityToFeatures,
    EntityValueRow,
    Feature as ServingProtoFeature,
    RequestedEntitiesMatrix,
    RequestedEntitiesMatrixRequest,
    RequestMetaData,
    StreamingAggregationFeature as ServingProtoStreamingAggregationFeature,
    StreamingFeature as ServingProtoStreamingFeature,
    StreamingV1Feature as ServingProtoStreamingV1Feature,
)
from _qwak_proto.qwak.feature_store.serving.serving_pb2_grpc import ServingServiceStub
from qwak.clients.administration.eco_system.client import EcosystemClient
from qwak.exceptions import QwakException
from qwak.feature_store._common.feature_set_utils import (
    BatchFeature,
    BatchFeatureV1,
    StreamingAggregationFeature,
    StreamingFeature,
    StreamingFeatureV1,
)
from qwak.feature_store.online import endpoint_utils
from qwak.feature_store.online.endpoint_utils import EndpointConfig
from qwak.inner.tool.grpc.grpc_tools import create_grpc_channel
from qwak.model.schema import ModelSchema
from qwak.model.schema_entities import BaseFeature, RequestInput

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    try:
        import pandas as pd
    except ImportError:
        pass


class OnlineClient:
    """
    Online Feature Serving client
    """

    def __init__(
        self,
        enable_ssl=True,
        endpoint_url: Optional[str] = None,
        metadata: Optional[Sequence[Tuple[str]]] = None,
    ):
        self._metadata = metadata

        options = (
            ("grpc.keepalive_timeout_ms", 1500),
            ("grpc.client_idle_timeout_ms", 60 * 1000),
        )

        if endpoint_url is None:
            user_context: AuthenticatedUnifiedUserContext = (
                EcosystemClient().get_authenticated_user_context().user
            )
            conf: EndpointConfig = endpoint_utils.get_config_by_account_type(
                user_context
            )
            endpoint_url = conf.endpoint_url
            if conf.enable_ssl is not None:
                enable_ssl = conf.enable_ssl

            if conf.metadata is not None:
                self._metadata = conf.metadata

        channel = create_grpc_channel(
            url=endpoint_url, enable_ssl=enable_ssl, options=options
        )

        self._serving_client: ServingServiceStub = ServingServiceStub(channel)

    @staticmethod
    def to_string_entities_values(values) -> List[str]:
        return [str(value) for value in values]

    def get_feature_values(
        self,
        schema: ModelSchema,
        df: "pd.DataFrame",
        model_name: str = "no-model",
        *,
        max_keys_per_request: int = 500,
    ) -> "pd.DataFrame":
        """
        :param schema: a ModelSchema object - defines the entities, features and prediction (irelevant in this case).
        :param df: a pandas data-frame with a column for each explicit feature needed
                         and a column for each entity key defined in the schema.
        :param model_name: the name of the model to retrieve the features for.
        :param max_keys_per_request: the maximum number of keys per request to the backend
        :return: a pandas data-frame - the feature values defined in the schema
                                       of the requested entities in the df.

        each row in the returned data-frame is constructed by retrieving the most recent requested feature values
        of the entity key(s) for the specific entity value(s) defined in the df.

        TODO: fix imports and align example
        Examples:
        >>> import pandas as pd
        >>> from qwak.feature_store import OnlineClient
        >>> from qwak.model.schema import (
        >>>     ModelSchema, FeatureStoreInput
        >>> )
        >>>
        >>> user_id = Entity(name='uuid', type=str)
        >>>
        >>> model_schema = ModelSchema(
        >>>     entities=[
        >>>         user_id
        >>>     ],
        >>>     inputs=[
        >>>         FeatureStoreInput(entity=user_id, name='user_purchases.number_of_purchases'),
        >>>         FeatureStoreInput(entity=user_id, name='user_purchases.avg_purchase_amount'),
        >>>     ],
        >>>     outputs=[
        >>>         InferenceOutput(name="score", type=float)
        >>>     ])
        >>>
        >>> online_client = OnlineClient()
        >>>
        >>> df = pd.DataFrame(columns=  ['uuid', 'explicit_feature_purchase_price'],
        >>>                   data   =  [ '1'  ,                22                ])
        >>>
        >>> user_features = online_client.get_feature_values(
        >>>                     model_schema,
        >>>                     df)
        >>>
        >>> print(user_features.head())
        >>>	#       user_purchases.number_of_purchases	user_purchases.avg_purchase_amount    otf_quad_price
        >>> # 0	                    76	                              4.796842                     484
        """
        try:
            import pandas as pd
        except ImportError:
            raise QwakException(
                "Missing Pandas dependency required for querying the online feature store"
            )

        (
            entity_features_compounds,
            entities_with_index,
            feature_set_names,
        ) = self._create_entity_and_features_sets(
            schema, pd.DataFrame(df.iloc[0]).transpose()
        )

        if not entity_features_compounds:
            df_result = pd.DataFrame(
                columns=[
                    input_field.name
                    for input_field in schema._inputs
                    if isinstance(input_field, BaseFeature)
                ]
            )

            return pd.concat([df, df_result], axis=1, join="inner")

        entities_to_features = []
        for entity_features_compound in entity_features_compounds.values():
            entities_to_features.append(
                EntityToFeatures(
                    features=entity_features_compound.features,
                    entity_name=entity_features_compound.entity_name,
                )
            )

        ordered_entities_tuple = sorted(
            entities_with_index, key=lambda entity: entity[1]
        )
        ordered_entities = [entity[0] for entity in ordered_entities_tuple]

        request_chunks: List[
            Tuple[RequestedEntitiesMatrix, pd.DataFrame]
        ] = OnlineClient._split_entities(
            entity_names=ordered_entities,
            population_df=df,
            max_entities_per_split=max_keys_per_request,
        )

        results: List[pd.DataFrame] = []
        for matrix, population_df in request_chunks:
            request = RequestedEntitiesMatrixRequest(
                entity_values_matrix=matrix,
                entities_to_features=entities_to_features,
                request_meta_data=RequestMetaData(
                    model_name=model_name,
                    feature_set_names=list(feature_set_names),
                    num_keys=len(population_df),
                ),
            )
            try:
                response_df_json, _ = self._serving_client.GetMultiFeatures.with_call(
                    request, metadata=self._metadata
                )
                features_df: pd.DataFrame = pd.read_json(
                    StringIO(response_df_json.pandas_df_as_json), orient="split"
                )
                results.append(
                    pd.concat(
                        [population_df.reset_index(drop=True), features_df],
                        axis=1,
                        join="inner",
                    )
                )
            except Exception as e:
                raise QwakException(
                    f"Failed to retrieve online feature values, got {repr(e)}"
                )

        res = pd.concat(results, axis=0)
        res.index = df.index
        return res

    @staticmethod
    def _split_entities(
        entity_names: List[str],
        population_df: "pd.DataFrame",
        max_entities_per_split: int,
    ) -> List[Tuple[RequestedEntitiesMatrix, "pd.DataFrame"]]:
        entity_values: List[List[str]] = population_df[entity_names].values.tolist()

        num_batches: int = math.ceil(len(entity_values) / max_entities_per_split)

        res: List[Tuple[RequestedEntitiesMatrix, pd.DataFrame]] = []
        for i in range(num_batches):
            curr_entities: List[List[str]] = entity_values[
                i * max_entities_per_split : (i + 1) * max_entities_per_split
            ]
            curr_population_df = population_df[
                i * max_entities_per_split : (i + 1) * max_entities_per_split
            ]
            curr_entity_rows: List[EntityValueRow] = [
                EntityValueRow(
                    index=idx, entity_values=OnlineClient.to_string_entities_values(v)
                )
                for idx, v in enumerate(curr_entities)
            ]

            res.append(
                (
                    RequestedEntitiesMatrix(
                        header=EntitiesHeader(entity_names=entity_names),
                        rows=curr_entity_rows,
                    ),
                    curr_population_df,
                )
            )
        return res

    @staticmethod
    def _create_entity_and_features_sets(schema: ModelSchema, df: "pd.DataFrame"):
        entity_features_compounds = {}
        list_of_df_columns = df.columns.to_list()
        entities_and_indexes = []
        feature_set_names = set()
        for entity in schema._entities:
            if entity.name not in df:
                logger.error(
                    f"Schema entity key '{entity.name}' does not exist in the request DataFrame"
                )
            else:
                entity_features_compounds[entity.name] = EntityFeaturesCompound(
                    entity.name, df[entity.name]
                )
                entities_and_indexes.append(
                    (entity.name, list_of_df_columns.index(entity.name))
                )

        for feature in [
            feature
            for feature in schema._inputs
            if not isinstance(feature, RequestInput)
        ]:
            if feature.entity.name not in entity_features_compounds:
                logger.info(
                    f"The entity: {feature.entity.name} of the Feature: {feature} does not exist in the entities list"
                )

            else:
                if isinstance(feature, BatchFeatureV1):
                    feature_proto = ServingProtoFeature(
                        batch_v1_feature=ServingProtoBatchV1Feature(name=feature.name)
                    )
                elif isinstance(feature, BatchFeature):
                    feature_proto = ServingProtoFeature(
                        batch_feature=ServingProtoBatchFeature(name=feature.name)
                    )
                elif isinstance(feature, StreamingFeature):
                    feature_proto = ServingProtoFeature(
                        streaming_feature=ServingProtoStreamingFeature(
                            name=feature.name
                        )
                    )
                elif isinstance(feature, StreamingFeatureV1):
                    feature_proto = ServingProtoFeature(
                        streaming_v1_feature=ServingProtoStreamingV1Feature(
                            name=feature.name
                        )
                    )
                elif isinstance(feature, StreamingAggregationFeature):
                    feature_proto = ServingProtoFeature(
                        streaming_aggregation=ServingProtoStreamingAggregationFeature(
                            name=feature.name
                        )
                    )
                else:
                    raise ValueError(
                        f"Not support type for feature extraction - {type(feature)}"
                    )

                entity_features_compounds[feature.entity.name].add_feature(
                    feature_proto
                )
                feature_set_names.add(feature.name.split(".")[0])

        return entity_features_compounds, entities_and_indexes, feature_set_names


class EntityFeaturesCompound:
    def __init__(self, entity_name, entity_value):
        self.entity_name = entity_name
        self.entity_value = entity_value
        self.features = []

    def add_feature(self, feature):
        self.features.append(feature)



from dataclasses import dataclass
from typing import List

from _qwak_proto.qwak.offline.serving.v1.feature_values_pb2 import (
    FeaturesetFeatures as ProtoFeaturesetFeatures,
)


@dataclass
class FeatureSetFeatures:
    feature_set_name: str
    feature_names: List[str]

    def to_proto(self) -> ProtoFeaturesetFeatures:
        return ProtoFeaturesetFeatures(
            featureset_name=self.feature_set_name,
            feature_names=self.feature_names,
        )

    @classmethod
    def from_proto(cls, featureset_features_proto: ProtoFeaturesetFeatures):
        return cls(
            feature_set_name=featureset_features_proto.feature_set_name,
            feature_names=featureset_features_proto.feature_names,
        )



from functools import lru_cache
from typing import Union

from qwak.llmops.generation.chat.openai.types.chat.chat_completion import ChatCompletion
from qwak.llmops.generation.streaming import ChatCompletionStream
from qwak.llmops.model.descriptor import ChatModelDescriptor, OpenAIChat
from qwak.llmops.prompt.chat.value import ChatPromptValue
from qwak.llmops.provider.openai.provider import OpenAIProvider


class ChatCompletionProvider:
    @staticmethod
    @lru_cache(maxsize=None)
    def _get_openai_provider():
        return OpenAIProvider()

    @staticmethod
    def invoke(
        chat_prompt_value: ChatPromptValue,
        chat_model_descriptor: ChatModelDescriptor,
        stream: bool = False,
    ) -> Union[ChatCompletion, ChatCompletionStream]:
        if isinstance(chat_model_descriptor, OpenAIChat):
            return ChatCompletionProvider._invoke_openai_chat(
                chat_prompt_value=chat_prompt_value,
                chat_model_descriptor=chat_model_descriptor,
                stream=stream,
            )
        else:
            raise ValueError("Can't invoke prompt and model combination!")

    @staticmethod
    def _invoke_openai_chat(
        chat_prompt_value: ChatPromptValue,
        chat_model_descriptor: OpenAIChat,
        stream: bool = False,
    ) -> Union[ChatCompletion, ChatCompletionStream]:
        return ChatCompletionProvider._get_openai_provider().create_chat_completion(
            chat_prompt_value=chat_prompt_value,
            chat_model_descriptor=chat_model_descriptor,
            stream=stream,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qwak/feature_store/platform/platform_details.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from _qwak_proto.qwak.feature_store.repository.common import platform_pb2 as qwak_dot_feature__store_dot_repository_dot_common_dot_platform__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2qwak/feature_store/platform/platform_details.proto\x12\x1bqwak.feature.store.platform\x1a\x33qwak/feature_store/repository/common/platform.proto\"\x9e\x01\n\x0fPlatformDetails\x12Q\n\x11platform_settings\x18\x01 \x01(\x0b\x32\x36.qwak.feature.store.repository.common.PlatformSettings\x12\x38\n\tauth_type\x18\x02 \x01(\x0b\x32%.qwak.feature.store.platform.AuthType\"&\n\x08\x41uthType\x12\x12\n\x08token_id\x18\x01 \x01(\tH\x00\x42\x06\n\x04typeB[\n&com.qwak.ai.feature.store.platform.apiP\x01Z/qwak/featurestore/features;featurestoreplatformb\x06proto3')



_PLATFORMDETAILS = DESCRIPTOR.message_types_by_name['PlatformDetails']
_AUTHTYPE = DESCRIPTOR.message_types_by_name['AuthType']
PlatformDetails = _reflection.GeneratedProtocolMessageType('PlatformDetails', (_message.Message,), {
  'DESCRIPTOR' : _PLATFORMDETAILS,
  '__module__' : 'qwak.feature_store.platform.platform_details_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.platform.PlatformDetails)
  })
_sym_db.RegisterMessage(PlatformDetails)

AuthType = _reflection.GeneratedProtocolMessageType('AuthType', (_message.Message,), {
  'DESCRIPTOR' : _AUTHTYPE,
  '__module__' : 'qwak.feature_store.platform.platform_details_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.platform.AuthType)
  })
_sym_db.RegisterMessage(AuthType)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n&com.qwak.ai.feature.store.platform.apiP\001Z/qwak/featurestore/features;featurestoreplatform'
  _PLATFORMDETAILS._serialized_start=137
  _PLATFORMDETAILS._serialized_end=295
  _AUTHTYPE._serialized_start=297
  _AUTHTYPE._serialized_end=335
# @@protoc_insertion_point(module_scope)



from abc import ABC, abstractmethod
from logging import Logger
from pathlib import Path
from typing import Any, Dict, List, Tuple, Optional

from marshmallow import RAISE, Schema
from marshmallow_dataclass import class_schema
from qwak.exceptions import QwakException, LoadConfigurationException
from yaml import SafeLoader, dump, load

from .utils import ConfigCliMap, rgetattr, rsetattr

ERROR_OPENING_YAML_FILE_MSG = "Error opening the config file"
ERROR_PARSING_YAML_FILE_MSG = "Error parsing the config file's yaml"
ERROR_IN_CONFIG_MSG = "Error loading the config"


class QwakConfigBase(ABC):
    """Base qwak config object."""

    @property
    @abstractmethod
    def _config_mapping(self) -> List[ConfigCliMap]:
        """Config mapping, Return a list of ConfigCliMap object in order to create the mapping.

        Returns:
            list: List of ConfigCliMap to apply.
        """
        pass

    def merge_cli_argument(
        self, sections: Tuple[str, ...] = (), **kwargs: Dict[str, Any]
    ):
        """Merge and validate cli arguments by supplied mapping.

        Args:
            sections: Sections to validate.
            **kwargs: argument from cli.

        Raises:
            QwakException: In case that the argument is not valid.
        """
        for prop_map in self._config_mapping:
            value = kwargs.get(prop_map.key)
            if value is not None:
                if isinstance(value, (list, tuple)):
                    new_value = list(rgetattr(self, prop_map.prop))
                    new_value.extend(value)
                    value = new_value
                rsetattr(self, prop_map.prop, value)
            if (
                not sections
                or any(
                    list(
                        map(
                            lambda section, _prop_map=prop_map: _prop_map.prop.startswith(
                                section
                            ),
                            sections,
                        )
                    )
                )
                or "." not in prop_map.prop
            ):
                config_value = rgetattr(self, prop_map.prop)
                if not prop_map.validation_func(config_value, prop_map.is_required):
                    raise QwakException(
                        f"{prop_map.key} argument contain invalid argument: "
                        f"{value or config_value}"
                    )
        self._post_merge_cli()

    @abstractmethod
    def _post_merge_cli(self):
        """Actions to perform after merging cli argument in to properties"""
        pass


class YamlConfigMixin(object):
    @classmethod
    def from_yaml(
        cls, yaml_path: str, unknown: str = RAISE, logger: Optional[Logger] = None
    ) -> Any:
        """Create instance of class from yaml and class scheme.

        Args:
            yaml_path (str): Yaml path.

        Returns:
            object: Instance of created class.
        """
        if not yaml_path:
            return cls()
        Schema.Meta.unknown = unknown
        schema = class_schema(cls)
        try:
            yaml_content = Path(yaml_path).read_text()
        except Exception as e:
            if logger:
                logger.error(f"Error opening the config file {yaml_path}")
            raise LoadConfigurationException(ERROR_OPENING_YAML_FILE_MSG) from e

        try:
            yaml_parsed = load(stream=yaml_content, Loader=SafeLoader)
        except Exception as e:
            if logger:
                logger.exception(ERROR_PARSING_YAML_FILE_MSG, e)
            raise LoadConfigurationException(ERROR_PARSING_YAML_FILE_MSG) from e

        try:
            return schema(unknown=unknown).load(yaml_parsed)
        except Exception as e:
            if logger:
                logger.exception(ERROR_IN_CONFIG_MSG, e)
            raise LoadConfigurationException(ERROR_IN_CONFIG_MSG) from e

    def to_yaml(self) -> str:
        """Convert class by scheme to yaml.

        Returns:
            str: Class as yaml string by scheme.
        """
        loaded_type = type(self)
        schema = class_schema(loaded_type)

        return dump(schema().dump(self))



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from _qwak_proto.qwak.projects import projects_pb2 as qwak_dot_projects_dot_projects__pb2


class ProjectsManagementServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateProject = channel.unary_unary(
                '/qwak.projects.management.ProjectsManagementService/CreateProject',
                request_serializer=qwak_dot_projects_dot_projects__pb2.CreateProjectRequest.SerializeToString,
                response_deserializer=qwak_dot_projects_dot_projects__pb2.CreateProjectResponse.FromString,
                )
        self.ListProjects = channel.unary_unary(
                '/qwak.projects.management.ProjectsManagementService/ListProjects',
                request_serializer=qwak_dot_projects_dot_projects__pb2.ListProjectsRequest.SerializeToString,
                response_deserializer=qwak_dot_projects_dot_projects__pb2.ListProjectsResponse.FromString,
                )
        self.GetProject = channel.unary_unary(
                '/qwak.projects.management.ProjectsManagementService/GetProject',
                request_serializer=qwak_dot_projects_dot_projects__pb2.GetProjectRequest.SerializeToString,
                response_deserializer=qwak_dot_projects_dot_projects__pb2.GetProjectResponse.FromString,
                )
        self.DeleteProject = channel.unary_unary(
                '/qwak.projects.management.ProjectsManagementService/DeleteProject',
                request_serializer=qwak_dot_projects_dot_projects__pb2.DeleteProjectRequest.SerializeToString,
                response_deserializer=qwak_dot_projects_dot_projects__pb2.DeleteProjectResponse.FromString,
                )
        self.UpdateProject = channel.unary_unary(
                '/qwak.projects.management.ProjectsManagementService/UpdateProject',
                request_serializer=qwak_dot_projects_dot_projects__pb2.UpdateProjectRequest.SerializeToString,
                response_deserializer=qwak_dot_projects_dot_projects__pb2.UpdateProjectResponse.FromString,
                )


class ProjectsManagementServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def CreateProject(self, request, context):
        """Create project
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListProjects(self, request, context):
        """List all projects
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetProject(self, request, context):
        """Get project
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteProject(self, request, context):
        """Delete project
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateProject(self, request, context):
        """Update project
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ProjectsManagementServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateProject': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateProject,
                    request_deserializer=qwak_dot_projects_dot_projects__pb2.CreateProjectRequest.FromString,
                    response_serializer=qwak_dot_projects_dot_projects__pb2.CreateProjectResponse.SerializeToString,
            ),
            'ListProjects': grpc.unary_unary_rpc_method_handler(
                    servicer.ListProjects,
                    request_deserializer=qwak_dot_projects_dot_projects__pb2.ListProjectsRequest.FromString,
                    response_serializer=qwak_dot_projects_dot_projects__pb2.ListProjectsResponse.SerializeToString,
            ),
            'GetProject': grpc.unary_unary_rpc_method_handler(
                    servicer.GetProject,
                    request_deserializer=qwak_dot_projects_dot_projects__pb2.GetProjectRequest.FromString,
                    response_serializer=qwak_dot_projects_dot_projects__pb2.GetProjectResponse.SerializeToString,
            ),
            'DeleteProject': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteProject,
                    request_deserializer=qwak_dot_projects_dot_projects__pb2.DeleteProjectRequest.FromString,
                    response_serializer=qwak_dot_projects_dot_projects__pb2.DeleteProjectResponse.SerializeToString,
            ),
            'UpdateProject': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateProject,
                    request_deserializer=qwak_dot_projects_dot_projects__pb2.UpdateProjectRequest.FromString,
                    response_serializer=qwak_dot_projects_dot_projects__pb2.UpdateProjectResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qwak.projects.management.ProjectsManagementService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ProjectsManagementService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def CreateProject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.projects.management.ProjectsManagementService/CreateProject',
            qwak_dot_projects_dot_projects__pb2.CreateProjectRequest.SerializeToString,
            qwak_dot_projects_dot_projects__pb2.CreateProjectResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListProjects(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.projects.management.ProjectsManagementService/ListProjects',
            qwak_dot_projects_dot_projects__pb2.ListProjectsRequest.SerializeToString,
            qwak_dot_projects_dot_projects__pb2.ListProjectsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetProject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.projects.management.ProjectsManagementService/GetProject',
            qwak_dot_projects_dot_projects__pb2.GetProjectRequest.SerializeToString,
            qwak_dot_projects_dot_projects__pb2.GetProjectResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteProject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.projects.management.ProjectsManagementService/DeleteProject',
            qwak_dot_projects_dot_projects__pb2.DeleteProjectRequest.SerializeToString,
            qwak_dot_projects_dot_projects__pb2.DeleteProjectResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateProject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.projects.management.ProjectsManagementService/UpdateProject',
            qwak_dot_projects_dot_projects__pb2.UpdateProjectRequest.SerializeToString,
            qwak_dot_projects_dot_projects__pb2.UpdateProjectResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _MetricType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _MetricTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_MetricType.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    METRIC_TYPE_NOT_VALID: _MetricType.ValueType  # 0
    METRIC_TYPE_CPU: _MetricType.ValueType  # 1
    METRIC_TYPE_MEMORY: _MetricType.ValueType  # 2
    METRIC_TYPE_LATENCY: _MetricType.ValueType  # 3
    METRIC_TYPE_GPU: _MetricType.ValueType  # 4
    METRIC_TYPE_ERROR_RATE: _MetricType.ValueType  # 5
    METRIC_TYPE_THROUGHPUT: _MetricType.ValueType  # 6

class MetricType(_MetricType, metaclass=_MetricTypeEnumTypeWrapper): ...

METRIC_TYPE_NOT_VALID: MetricType.ValueType  # 0
METRIC_TYPE_CPU: MetricType.ValueType  # 1
METRIC_TYPE_MEMORY: MetricType.ValueType  # 2
METRIC_TYPE_LATENCY: MetricType.ValueType  # 3
METRIC_TYPE_GPU: MetricType.ValueType  # 4
METRIC_TYPE_ERROR_RATE: MetricType.ValueType  # 5
METRIC_TYPE_THROUGHPUT: MetricType.ValueType  # 6
global___MetricType = MetricType

class _AggregationType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _AggregationTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_AggregationType.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    AGGREGATION_TYPE_NOT_VALID: _AggregationType.ValueType  # 0
    AGGREGATION_TYPE_MIN: _AggregationType.ValueType  # 1
    AGGREGATION_TYPE_MAX: _AggregationType.ValueType  # 2
    AGGREGATION_TYPE_AVERAGE: _AggregationType.ValueType  # 3
    AGGREGATION_TYPE_SUM: _AggregationType.ValueType  # 4
    AGGREGATION_TYPE_P50: _AggregationType.ValueType  # 5
    AGGREGATION_TYPE_P90: _AggregationType.ValueType  # 6
    AGGREGATION_TYPE_P95: _AggregationType.ValueType  # 7
    AGGREGATION_TYPE_P99: _AggregationType.ValueType  # 8

class AggregationType(_AggregationType, metaclass=_AggregationTypeEnumTypeWrapper): ...

AGGREGATION_TYPE_NOT_VALID: AggregationType.ValueType  # 0
AGGREGATION_TYPE_MIN: AggregationType.ValueType  # 1
AGGREGATION_TYPE_MAX: AggregationType.ValueType  # 2
AGGREGATION_TYPE_AVERAGE: AggregationType.ValueType  # 3
AGGREGATION_TYPE_SUM: AggregationType.ValueType  # 4
AGGREGATION_TYPE_P50: AggregationType.ValueType  # 5
AGGREGATION_TYPE_P90: AggregationType.ValueType  # 6
AGGREGATION_TYPE_P95: AggregationType.ValueType  # 7
AGGREGATION_TYPE_P99: AggregationType.ValueType  # 8
global___AggregationType = AggregationType

class AutoScalingConfig(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MIN_REPLICA_COUNT_FIELD_NUMBER: builtins.int
    MAX_REPLICA_COUNT_FIELD_NUMBER: builtins.int
    POLLING_INTERVAL_FIELD_NUMBER: builtins.int
    COOL_DOWN_PERIOD_FIELD_NUMBER: builtins.int
    TRIGGERS_FIELD_NUMBER: builtins.int
    min_replica_count: builtins.int
    """The minimum number of replicas will scale the resource down to"""
    max_replica_count: builtins.int
    """The maximum number of replicas of the target resource"""
    polling_interval: builtins.int
    """This is the interval to check each trigger on. By default it's every 30 seconds."""
    cool_down_period: builtins.int
    """The period to wait after the last trigger reported active before scaling the resource back to 0. By default it’s 5 minutes (300 seconds)."""
    @property
    def triggers(self) -> global___Triggers: ...
    def __init__(
        self,
        *,
        min_replica_count: builtins.int = ...,
        max_replica_count: builtins.int = ...,
        polling_interval: builtins.int = ...,
        cool_down_period: builtins.int = ...,
        triggers: global___Triggers | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["triggers", b"triggers"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["cool_down_period", b"cool_down_period", "max_replica_count", b"max_replica_count", "min_replica_count", b"min_replica_count", "polling_interval", b"polling_interval", "triggers", b"triggers"]) -> None: ...

global___AutoScalingConfig = AutoScalingConfig

class Triggers(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TRIGGERS_FIELD_NUMBER: builtins.int
    @property
    def triggers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Trigger]:
        """List of triggers to activate scaling(up/down) of the target resource based on"""
    def __init__(
        self,
        *,
        triggers: collections.abc.Iterable[global___Trigger] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["triggers", b"triggers"]) -> None: ...

global___Triggers = Triggers

class Trigger(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PROMETHEUS_TRIGGER_FIELD_NUMBER: builtins.int
    @property
    def prometheus_trigger(self) -> global___PrometheusTrigger:
        """Prometheus trigger type"""
    def __init__(
        self,
        *,
        prometheus_trigger: global___PrometheusTrigger | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["prometheus_trigger", b"prometheus_trigger", "trigger_type", b"trigger_type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["prometheus_trigger", b"prometheus_trigger", "trigger_type", b"trigger_type"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["trigger_type", b"trigger_type"]) -> typing_extensions.Literal["prometheus_trigger"] | None: ...

global___Trigger = Trigger

class PrometheusTrigger(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    QUERY_SPEC_FIELD_NUMBER: builtins.int
    THRESHOLD_FIELD_NUMBER: builtins.int
    @property
    def query_spec(self) -> global___QuerySpec:
        """Query specification"""
    threshold: builtins.int
    """Value to start scaling for"""
    def __init__(
        self,
        *,
        query_spec: global___QuerySpec | None = ...,
        threshold: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["query_spec", b"query_spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["query_spec", b"query_spec", "threshold", b"threshold"]) -> None: ...

global___PrometheusTrigger = PrometheusTrigger

class QuerySpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    METRIC_TYPE_FIELD_NUMBER: builtins.int
    AGGREGATION_TYPE_FIELD_NUMBER: builtins.int
    TIME_PERIOD_FIELD_NUMBER: builtins.int
    ERROR_CODE_FIELD_NUMBER: builtins.int
    metric_type: global___MetricType.ValueType
    """The type of the metric"""
    aggregation_type: global___AggregationType.ValueType
    """The type of the aggregation"""
    time_period: builtins.int
    """The period to run the query based on"""
    error_code: builtins.str
    """The error code to scale by in case of MetricType.METRIC_TYPE_ERROR_RATE - default is ".*" """
    def __init__(
        self,
        *,
        metric_type: global___MetricType.ValueType = ...,
        aggregation_type: global___AggregationType.ValueType = ...,
        time_period: builtins.int = ...,
        error_code: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["aggregation_type", b"aggregation_type", "error_code", b"error_code", "metric_type", b"metric_type", "time_period", b"time_period"]) -> None: ...

global___QuerySpec = QuerySpec



import json
from typing import List, Optional, Union

from qwak.exceptions import QwakHTTPException

from .string_input import StringInput

try:
    import pandas
    from bentoml.adapters import DataframeInput as BentoMLDataframeInput
except ImportError:
    pass


class DataframeInput(StringInput):
    def __init__(
        self,
        input_orient: Optional[str] = None,
        **base_kwargs,
    ):
        self.input_orient = input_orient
        super().__init__(**base_kwargs)

        # Verify pandas imported properly and retry import if it has failed initially
        try:
            import pandas  # noqa: F401
        except ImportError:
            raise ImportError("Pandas package is required to use DataframeInput.")

    def extract_user_func_args(
        self,
        data: Union[
            str, List
        ],  # it's a list of InferenceTasks, but we don't want to fail when bentoml is not installed
    ) -> "pandas.DataFrame":
        try:
            if isinstance(data, list):
                adapter = BentoMLDataframeInput(orient=self.input_orient)
                # because parsing is deferred to bentoml, it will also set the batch property of a task
                (df,) = adapter.extract_user_func_args(data)
                return df
            else:
                return pandas.DataFrame.from_dict(json.loads(data), orient="columns")
        except Exception as e:
            raise QwakHTTPException(
                status_code=400,
                message=f"Error loading DataFrame input: {e}.",
            )



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Dict, List

from _qwak_proto.qwak.build.v1.build_pb2 import Audit as AuditProto
from _qwak_proto.qwak.build.v1.build_pb2 import Build as BuildProto
from _qwak_proto.qwak.builds.builds_pb2 import Build as BuildsManagementBuild
from google.protobuf.timestamp_pb2 import Timestamp


class BuildStatus(Enum):
    INVALID = 0
    IN_PROGRESS = 1
    SUCCESSFUL = 2
    FAILED = 3
    REMOTE_BUILD_INITIALIZING = 4
    REMOTE_BUILD_CANCELLED = 5
    REMOTE_BUILD_TIMED_OUT = 6
    REMOTE_BUILD_UNKNOWN = 7
    SYNCING_ENVIRONMENTS = 8
    FINISHED_SYNCING = 9


@dataclass
class Build:
    model_id: str = field(default=None)
    build_id: str = field(default=None)
    commit_id: str = field(default=None)
    build_status: BuildStatus = field(default=BuildStatus.INVALID)
    tags: List[str] = field(default_factory=list)
    parameters: Dict[str, str] = field(default_factory=dict)
    metrics: Dict[str, float] = field(default_factory=dict)
    created_by: str = field(default="")
    created_at: datetime = field(default=datetime.now())
    last_modified_by: str = field(default="")
    last_modified_at: datetime = field(default=datetime.now())

    def to_proto(self):
        timestamp = Timestamp()
        return BuildProto(
            buildId=self.build_id,
            commitId=self.commit_id,
            build_status=self.build_status.value,
            tags=self.tags,
            params=self.parameters,
            metrics=self.metrics,
            audit=AuditProto(
                created_by=self.created_by,
                created_at=timestamp.FromDatetime(self.created_at),
                last_modified_by=self.last_modified_by,
                last_modified_at=timestamp.FromDatetime(self.last_modified_at),
            ),
        )

    @staticmethod
    def from_proto(build_proto: BuildProto):
        return Build(
            build_id=build_proto.buildId,
            commit_id=build_proto.commitId,
            build_status=BuildStatus(build_proto.build_status),
            tags=build_proto.tags,
            parameters=build_proto.params,
            metrics=build_proto.metrics,
            created_by=build_proto.audit.created_by,
            created_at=datetime.fromtimestamp(
                build_proto.audit.created_at.seconds
                + build_proto.audit.created_at.nanos / 1e9
            ),
            last_modified_by=build_proto.audit.last_modified_by,
            last_modified_at=datetime.fromtimestamp(
                build_proto.audit.last_modified_at.seconds
                + build_proto.audit.last_modified_at.nanos / 1e9
            ),
        )

    @staticmethod
    def from_builds_management(build_proto: BuildsManagementBuild):
        return Build(
            model_id=build_proto.build_spec.model_id,
            build_id=build_proto.build_spec.build_id,
            commit_id=build_proto.build_spec.commit_id,
            build_status=BuildStatus(build_proto.build_status),
            tags=build_proto.build_spec.tags,
            parameters=build_proto.build_spec.experiment_tracking_values.params,
            metrics=build_proto.build_spec.experiment_tracking_values.metrics,
            created_by=build_proto.created_by,
            created_at=datetime.fromtimestamp(
                build_proto.created_at.seconds + build_proto.created_at.nanos / 1e9
            ),
            last_modified_by=build_proto.last_modified_by,
            last_modified_at=datetime.fromtimestamp(
                build_proto.last_modified_at.seconds
                + build_proto.last_modified_at.nanos / 1e9
            ),
        )



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class GetDataSourceSourceCodeDownloadURLRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATA_SOURCE_NAME_FIELD_NUMBER: builtins.int
    data_source_name: builtins.str
    def __init__(
        self,
        *,
        data_source_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["data_source_name", b"data_source_name"]) -> None: ...

global___GetDataSourceSourceCodeDownloadURLRequest = GetDataSourceSourceCodeDownloadURLRequest

class GetDataSourceSourceCodeDownloadURLResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DOWNLOAD_URL_FIELD_NUMBER: builtins.int
    download_url: builtins.str
    def __init__(
        self,
        *,
        download_url: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["download_url", b"download_url"]) -> None: ...

global___GetDataSourceSourceCodeDownloadURLResponse = GetDataSourceSourceCodeDownloadURLResponse

class GetDataSourceSourceCodeUploadRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATA_SOURCE_NAME_FIELD_NUMBER: builtins.int
    data_source_name: builtins.str
    def __init__(
        self,
        *,
        data_source_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["data_source_name", b"data_source_name"]) -> None: ...

global___GetDataSourceSourceCodeUploadRequest = GetDataSourceSourceCodeUploadRequest

class GetDataSourceSourceCodeUploadResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    UPLOAD_URL_FIELD_NUMBER: builtins.int
    upload_url: builtins.str
    def __init__(
        self,
        *,
        upload_url: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["upload_url", b"upload_url"]) -> None: ...

global___GetDataSourceSourceCodeUploadResponse = GetDataSourceSourceCodeUploadResponse
